local M = {}
M.type = "auxiliary"
M.relevantDevice = nil
M.defaultOrder = 1000

local constants = {
    -- Conversion rates
    mpsToMph = 2.23694,
    hpToKw = 0.7457,
    kwToHp = 1.341021,
    torqueToHpCoef = 7120.54016,
    kWhToJ = 3600000,
    sToH = 1.0 / 3600,
    avToRPM = 9.549296596425384,
    rpmToAV = 1.0 / 9.549296596425384,
}

-- lib requires
local mathutil = require("mathutil")
local latch = require("latch")
local pid = require("pid")
local debugutil = require("debugutil")("hybrid", true)
local engineutil = require("engineUtil")
local simpleCurve = require("simplecurve")

local min = math.min
local max = math.max
local abs = math.abs
local maprange = mathutil.maprange
local clamp = mathutil.clamp
local blend = mathutil.blend
local round = mathutil.round
local sign = mathutil.sign
local fnum = debugutil.formatNumber

local engine = nil
local gearbox = nil
local battery = nil

local soundSources = {
    mg1Whine = nil,
    mg2Whine = nil,
    motorRamp = nil,
}

-- Calculated/initial values
local initialWheelBrakeSplit = {}
local torqueData = {
    maxElectricTorque = 0,
    fullBatteryElectricTorque = 0,
    maxEngineTorque = 0,
    maxCombinedTorque = 0,
    engineIdleTorque = 0,
    engineMaxTorqueRpm = 0,
    engineTorqueCurve = {},
}
local engineTorqueData

-- Quick access upvalues
local engineOn = false
local watertemp = 0
local speedMph = 0
local batteryLevel = 1
local totalPowerOutKw = 0
local throttle = 0
local throttleAggressiveness = 0
local gear = "P"
local brakeInput = 0

-- Other state variables/upvalues
local triggerEngineDemandReason = "none"
local engineDemandReason = "none"
local lastEngineDemandedOn = false
local miscEngineDemandReason = "none"
local miscEngineDemand = false
local waterTempEngineDemand = false
local state = "init"
local nextState = nil
local lastState = nil
local timeSincePowerOn = 0
local engineRunTimeSincePowerOn = 0
local timeSinceEngineStart = 5
local timeSinceEngineCanShutOff = 0
local timeSinceAppliedBrakes = 0
local ecvtDriveMode = 0
local startingTime = 0
local isInChargeup = false
local lastThrottle = 0
local chargeThrottle = 0
local disableRegen = false
local timeSinceWheelSlip = 0
local timeSinceInit = 0
local mg2TorqueAdjustment = 0
local engineDeload = 0
local desiredRegen = 0
local availableMG1Power = 0
local availableMG2Power = 0
local targetEngineRpm = 0
local availablePriorityEvPower = 0
local availableSecondaryEvPower = 0
local powerGeneratedFromEngine = 0
local maxBatteryPower = 0
local availableBatteryPower = 0
local batteryWorkingRatio = 0
local requestedChargePower = 0
local forcedChargeMode = false
local totalPossibleBrakeTorque = 0
local totalPropulsedBrakeTorque = 0
local totalUnpropulsedBrakeTorque = 0
local currentBrakingCoef = 0
local engineStartAttempts = 0
local shownDisabledMessage = false
local gearLockedInPark = false
local lastAverageWheelSpeed = 0
    
local function getProportionalBatteryRatio(minLevelBias, maxLevelBias)
    minLevelBias = minLevelBias or 0
    maxLevelBias = maxLevelBias or 0

    local minPowerLevelThreshold = blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, minLevelBias)
    local maxPowerLevelThreshold = blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 1 - maxLevelBias)

    return maprange(batteryLevel, minPowerLevelThreshold, maxPowerLevelThreshold)
end

local function calculateElectricLimits(dt)
    -- Calculate the coefficient of battery charge left (more power from the battery
    -- can be used at low state-of-charge in EV mode)
    if state == "evmode" then
        batteryWorkingRatio = maprange(batteryLevel, 0.15, M.maxBatteryPowerLevel)
    else
        batteryWorkingRatio = getProportionalBatteryRatio()
    end
    
    -- Calculate battery power limits
    local availableBatteryPowerCoef
    
    if state == "startup" then
        availableBatteryPowerCoef = maprange(batteryLevel, 0, 0.05)
    else
        availableBatteryPowerCoef = maprange(batteryWorkingRatio, 0, 0.05)
    end
    
    maxBatteryPower = ((battery.jbeamData.maxPowerOutput or 1e5) - M.idlePowerDraw)
    availableBatteryPower = availableBatteryPowerCoef * maxBatteryPower
    
    -- Figure out some general values about what the electric subsystem can currently provide
    if state == "evmode" then
        availablePriorityEvPower = availableBatteryPower
        availableSecondaryEvPower = availableBatteryPower
        powerGeneratedFromEngine = 0
    else
        local priorityPowerSpeed
        local priorityPowerBatteryLevel = M.maxPriorityPowerBatteryLevel:get(batteryWorkingRatio * 100)
        
        if engineOn then
            priorityPowerSpeed = M.maxEngineAssistPriorityPowerSpeed:get(speedMph)
        else
            priorityPowerSpeed = M.maxEvOnlyPriorityPowerSpeed:get(speedMph)
        end
        
        powerGeneratedFromEngine = max(0, -gearbox.motorGenerator1Power)
        availablePriorityEvPower = mathutil.minall(availableBatteryPower, priorityPowerSpeed, priorityPowerBatteryLevel)
        availableSecondaryEvPower = powerGeneratedFromEngine + availableBatteryPower
        --availableSecondaryEvPower = availableBatteryPower
        
        if engineOn then
            availableSecondaryEvPower = min(M.maxEngineAssistSecondaryPower, availableSecondaryEvPower)
        end
    end
end

local function calculateTorqueStatistics()
    -- Calculate some torque performance statistics
        
    -- Raw max torques from gearbox and engine
    local maxFinalMG2Torque = gearbox:getMaxMG2Torque(availablePriorityEvPower + availableSecondaryEvPower) * gearbox.finalMG2GearRatio
    local maxEngineTorque = engine.torqueData.maxTorque
    local maxSunGearTorque = gearbox:getMaxMG1Torque(true)
    local maxEngineTorqueToCarrier = maxEngineTorque * gearbox.inputGearRatio
    local maxEngineTorqueToSun = abs(gearbox:getSunTorqueFromCarrierTorque(maxEngineTorqueToCarrier))
    local maxUsableEngineTorqueToSun = min(maxEngineTorqueToSun, maxSunGearTorque)
    local maxEngineTorqueToOutput = abs(gearbox:getRingTorqueFromSunTorque(maxUsableEngineTorqueToSun)) * gearbox.finalRingGearRatio
    
    torqueData.maxEngineTorque = maxEngineTorqueToOutput
    
    if state == "evmode" then
        local maxFinalMG1Torque = gearbox:getRingTorqueFromSunTorque(gearbox:getMaxMG1Torque()) * gearbox.finalRingGearRatio
        
        torqueData.maxElectricTorque = maxFinalMG2Torque + maxFinalMG1Torque
        torqueData.fullBatteryElectricTorque = torqueData.maxElectricTorque
        torqueData.maxCombinedTorque = torqueData.maxElectricTorque
    else
        -- Figure out compound torques
        
        torqueData.maxElectricTorque = maxFinalMG2Torque
        torqueData.fullBatteryElectricTorque = gearbox:getMaxMG2Torque(maxBatteryPower) * gearbox.finalMG2GearRatio
        torqueData.maxCombinedTorque = torqueData.maxEngineTorque + torqueData.maxElectricTorque
    end
        
    if type(M.maxOutputTorque) == "number" then
        local vehiclePitchDegrees = max(0, obj:getDirectionVector().z * 180 / math.pi)
        local torqueBoost = min(M.maxTorqueBoost, M.maxTorqueBoostPerDegree * vehiclePitchDegrees)
        local maxOutputTorque = M.maxOutputTorque + torqueBoost
        
        torqueData.maxCombinedTorque = min(maxOutputTorque, torqueData.maxCombinedTorque)
    end
    
    -- Figure out engine torque characteristics
    local curve = engine.torqueData.curves[engine.torqueData.finalCurveName]
    local idleTorque = curve.torque[engine.idleRPM]
    
    torqueData.engineIdleTorque = idleTorque
    torqueData.engineMaxTorqueRpm = engine.torqueData.maxTorqueRPM
    torqueData.engineTorqueCurve = curve.torque
end

local function getTorqueData()
    return torqueData
end

local function printTorqueData()
    local curTorqueData = getTorqueData()
    
    debugutil.log("I", "Max Electric Torque: " .. fnum(curTorqueData.maxElectricTorque))
    debugutil.log("I", "Max Full-Battery Electric Torque: " .. fnum(curTorqueData.fullBatteryElectricTorque))
    debugutil.log("I", "Max Usable Engine Torque: " .. fnum(curTorqueData.maxEngineTorque))
    debugutil.log("I", "Max Optimal Total Torque: " .. fnum(curTorqueData.maxCombinedTorque))
    debugutil.log("I", "Max Engine Idle Torque: " .. fnum(curTorqueData.engineIdleTorque))
    debugutil.log("I", "Engine Max Torque RPM: " .. fnum(curTorqueData.engineMaxTorqueRpm))
end

local function requestEngine(reason)
    if miscEngineDemand then return end
    
    miscEngineDemandReason = reason
    miscEngineDemand = true
end

local function isEngineDemandedOn()
    local inputThrottle = electrics.values.throttleOverride or input.throttle
    local minRunTime
    
    if (engineOn or timeSincePowerOn > M.vehicleStartupEngineDelay) and engineRunTimeSincePowerOn < M.vehicleStartupEngineRuntime then
        engineDemandReason = "vehicle startup"
        return true
    end
    
    if watertemp < M.minCoolantTempWarmedUp then
        minRunTime = M.engineMinRunTime * 2
    else
        minRunTime = M.engineMinRunTime
    end

    if state == "chargemode" then
        engineDemandReason = "charge mode"
        return true
    end
    if not M.enableAutoStop then
        engineDemandReason = "auto stop disabled"
        return true
    end
    if watertemp < M.minCoolantTemp or waterTempEngineDemand then
        engineDemandReason = "temp"
        waterTempEngineDemand = true
        return true
    else
        
    end -- engine must be warmed up
    if miscEngineDemand then
        engineDemandReason = miscEngineDemandReason
        return true
    end
    
    if gear == "R" then--and timeSinceEngineStart >= minRunTime then
        -- After this point, if in R and battery >= 0.4, force engine off
        engineDemandReason = "none (forced off due to reverse)"
        return false
    end
    
    if speedMph > M.maxEvOnlySpeed then
        engineDemandReason = "speed"
        return true
    end -- too fast
    if obj:getDirectionVector().z > M.maxEvOnlyPitch then
        engineDemandReason = "vehicle pitch"
        return true
    end -- pitched too high up (hill start)
    if M.latchBatteryChargeUp:test(batteryWorkingRatio) then
        engineDemandReason = "battery level"
        return true
    end -- battery too low (first check; doesn't trigger chargeup)
    if batteryLevel < M.minBatteryPowerLevel or isInChargeup then
        engineDemandReason = "battery level"
        isInChargeup = true
        return true
    end -- battery too low (second check; triggers chargeup)
    if throttleAggressiveness > M.maxEvOnlyThrottleAggressiveness then
        engineDemandReason = "throttle"
        return true
    end
    
    if engineOn and (
        not (electrics.values.parkingbrake == 1 or currentBrakingCoef >= M.autoStopBrakeThreshold) or
        timeSinceAppliedBrakes < M.autoStopBrakeTime
    ) then
        -- this check only applies when not in P
        if gear ~= "P" then
            engineDemandReason = "brakes"
            return true
        end
    end

    -- Easy checks are done; dynamic checks now

    -- This is the percentage of the max EV-only kW output that's available to us,
    -- based on battery level

    if engineOn then
        if inputThrottle > M.engineShutOffThrottleThreshold then
            engineDemandReason = "accelerating with engine on"
            return true
        end -- once engine is on, must coast again to shut off
    else
        if gear ~= "R" and gearbox.motorGenerator2Power > M.maxEvOnlyPriorityPowerSpeed:get(speedMph) then
            engineDemandReason = "priority power exceeded"
            return true
        end
    end

    -- Put this last so it doesn't override other reasons
    if (state == "startup" or state == "running") and timeSinceEngineStart < minRunTime then
        engineDemandReason = "time since start"
        return true
    end -- engine has to run for at least 4 seconds

    engineDemandReason = "none"
    return false -- by default, the engine is off
end

local function updateValues(dt)
    if not gearbox or not engine or not battery then
        return
    end
    
    local throttleAggressivenessInst = (input.throttle - lastThrottle) / dt
    
    miscEngineDemandReason = "none"
    miscEngineDemand = false
    lastThrottle = input.throttle
    engineOn = not (engine.isStalled or engine.isDisabled or engine.ignitionCoef < 0.01 or electrics.values.rpm < engine.idleRPM / 2)
    watertemp = electrics.values.watertemp
    speedMph = electrics.values.wheelspeed * constants.mpsToMph
    batteryLevel = battery.remainingRatio
    gear = electrics.values.gear
    mg2TorqueAdjustment = 0
    brakeInput = input.brake
    
    -- Rolling average
    throttleAggressiveness = throttleAggressiveness + (throttleAggressivenessInst - throttleAggressiveness) / 10
    
    local throttleRate = maprange(
        abs(throttleAggressiveness),
        M.minThrottleAggressivenessRamp,
        M.maxThrottleAggressivenessRamp,
        M.minThrottleAggressivenessFactor,
        M.maxThrottleAggressivenessFactor
    )
    
    throttle = electrics.values.throttleOverride or M.throttleSmoother:get(input.throttle, dt * throttleRate)

    if engineOn then
        timeSinceEngineStart = timeSinceEngineStart + dt
        
        if watertemp >= M.minCoolantTemp + 2.5 then
            waterTempEngineDemand = false
        end
    else
        timeSinceEngineStart = 0
    end
    
    if M.powerOn then
        timeSincePowerOn = timeSincePowerOn + dt
    else
        timeSincePowerOn = 0
    end
    
    if M.autoStopBrakeThreshold > 0 and (currentBrakingCoef >= M.autoStopBrakeThreshold or electrics.values.parkingbrake == 1) and speedMph <= M.maxEvOnlySpeed then
        timeSinceAppliedBrakes = timeSinceAppliedBrakes + dt
    else
        timeSinceAppliedBrakes = 0
    end

    if isInChargeup and batteryLevel >= blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 0.2) then
        isInChargeup = false
    end
    
    debugutil.putValue("Throttle Rate", throttleRate)
    debugutil.putValue("Engine Run Time", timeSinceEngineStart)
end

local function stopEngine()
    engine.ignitionCoef = 0 -- Kill engine
    engine.isDisabled = true -- Prevent the user from manually restarting the engine
    electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)

    nextState = "stopped"
end

local function resumeNormalState()
    if engineOn then
        nextState = "running"
    else
        nextState = "stopped"
    end
end

local function updateState(dt)
    ---------------------------
    -- BEGIN ALL-STATE LOGIC --
    ---------------------------
    
    local engineDemandedOn = isEngineDemandedOn()

    if engineDemandedOn and not lastEngineDemandedOn then
        triggerEngineDemandReason = engineDemandReason
    elseif not engineDemandedOn and state ~= "startup" then
        triggerEngineDemandReason = "none"
    end
    
    lastEngineDemandedOn = engineDemandedOn
    
    debugutil.putValue("Power", M.powerOn and "On" or "Off")
    debugutil.putValue("Coolant Temp.", electrics.values.watertemp)
    debugutil.putValue("Engine demanded on", tostring(engineDemandedOn))
    debugutil.putValue("Engine on reason", engineDemandReason)
    debugutil.putValue("Engine triggered on by", triggerEngineDemandReason)
    debugutil.putValue("State", state)

    -- keep things like daytime-running-lights on
    electrics.values.ignition = M.powerOn and 1 or 0
    electrics.values.running = M.powerOn and 1 or 0
    
    if batteryLevel < M.criticalBatteryPowerLevel and M.hasAuxiliaryStarterMotor then
        -- Fall back to 12V starter
        
        engine.starterDisabled = false
    else
        -- prevent engine starter from working (since the eCVT starts it)
        
        engine:deactivateStarter()
        engine.starterDisabled = true
    end
    
    if not M.powerSwitch then
        electrics.values.chargeMode = 0
        electrics.values.evMode = 0
        stopEngine()
        nextState = "poweroff"
    end
    
    electrics.values.battery = batteryLevel < M.criticalBatteryPowerLevel
    
    debugutil.putValue("Combustion Torque", engine.combustionTorque)

    --------------------------------
    -- BEGIN STATE-SPECIFIC LOGIC --
    --------------------------------
    
    if state == "init" then
        timeSinceInit = timeSinceInit + dt
        
        if lastState ~= "init" then
            gearbox:setInputDisengaged(false)
            electrics.values.checkengine = true
            electrics.values.hazard = true
            electrics.values.battery = true
            isInChargeup = false
            engine:enable()
            
            -- Reset all the stateful things
            M.throttleSmoother:reset()
            M.throttleLimiterSmoother:reset()
            M.engineStarterSmoother:reset()
            M.chargePowerSmoother:reset()
            M.mg1MaxPowerSmoother:reset()
            M.mg2MaxPowerSmoother:reset()
            M.driveThrottlePid:reset()
            M.chargeThrottlePid:reset()
        end
        
        if engineStartAttempts < 3 then
            if timeSinceInit >= M.initializeTime then
                if not M.powerOn then
                    M.powerOn = true
                end
                
                electrics.values.evMode = input.evMode or 0
                gearLockedInPark = false
                
                if M.startEvMode or M.forceEvMode then
                    nextState = "evmode"
                elseif batteryLevel < M.criticalBatteryPowerLevel then
                    -- attempt to start engine
                    
                    if engineOn and electrics.values.rpm >= engine.idleRPM - 100 then
                        if gear == "P" and timeSinceEngineStart > M.engineStartupIdleTime then
                            nextState = "chargemode"
                            forcedChargeMode = true
                            electrics.values.chargeMode = 1
                            gui.message({txt = "Battery level critical...entering recovery charge-up mode"}, 5, "vehicle.hybrid.chargeMode")
                        end
                    elseif M.hasAuxiliaryStarterMotor then
                        if engine.starterEngagedCoef < 1 then
                            engine.ignitionCoef = 0
                            engine.isDisabled = false
                            engine.starterDisabled = false
                            startingTime = 0
                            timeSinceInit = 0
                            engine:activateStarter()
                        else
                            startingTime = startingTime + dt
                            
                            if startingTime > 3 then
                                -- Something's not right; retry at init
                                
                                engine.ignitionCoef = 0
                                engineStartAttempts = engineStartAttempts + 1
                                engine:deactivateStarter()
                                timeSinceInit = 0
                            end
                        end
                    else
                        resumeNormalState()
                        engineRunTimeSincePowerOn = 0.0
                    end
                else
                    if engineStartAttempts > 0 then
                        -- immediately try and restart the engine
                        
                        nextState = "startup"
                        timeSinceEngineStart = 0
                        startingTime = 0
                    else
                        resumeNormalState()
                        engineRunTimeSincePowerOn = 0.0
                    end
                end
            else
                -- Prevent shifting while initializing
                
                if gear == "P" then
                    gearLockedInPark = true
                end
                
                if gearLockedInPark and gear ~= "P" then
                    controller.mainController.shiftToGearIndex(1)
                end
            end
        else
            if not shownDisabledMessage then
                gui.message({txt = "Engine failed to start. Vehicle is disabled."}, 5, "vehicle.hybrid.general")
                shownDisabledMessage = true
                engine:disable()
            end
        end
    elseif state == "evmode" then
        engine.ignitionCoef = 0 -- Kill engine
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        gearbox:setInputDisengaged(true)
        
        if not M.forceEvMode and electrics.values.chargeMode == 1 then
            if gear == "P" then
                electrics.values.evMode = 0
            else
                gui.message({txt = "Charge mode can only be enabled in P gear"}, 5, "vehicle.hybrid.chargeMode")
                electrics.values.chargeMode = 0
            end
        end

        if electrics.values.evMode ~= 1 then
            if M.forceEvMode then
                electrics.values.evMode = 1
            else
                gui.message({txt = "EV mode disabled"}, 2, "vehicle.hybrid.evMode")
                gearbox:setInputDisengaged(false)
                resumeNormalState()
            end
        end
        
        if batteryWorkingRatio < 0.025 and not M.forceEvMode then
            electrics.values.evMode = 0
            gui.message({txt = "Battery too low; EV mode disabled"}, 2, "vehicle.hybrid.evMode")
            gearbox:setInputDisengaged(false)
            resumeNormalState()
        end
    elseif state == "chargemode" then
        local inverseRelativePowerLevel = maprange(batteryWorkingRatio, 0.95, 1.0, 1.0, 0.05)
        local maxFastChargePower = M.maxFastChargePower or gearbox.mg1PowerRating
        local curGenOutput = -min(totalPowerOutKw, 0)
        local targGenOutput = inverseRelativePowerLevel * maxFastChargePower

        if lastState ~= state then
            M.chargeThrottlePid:reset()
        end
        
        chargeThrottle = clamp(M.chargeThrottlePid:get(targGenOutput, curGenOutput, dt), 0, 1)

        if forcedChargeMode then
            if gear ~= "P" then
                gui.message({txt = "Cannot cancel recovery charge-up. Please wait for battery to charge."}, 5, "vehicle.hybrid.chargeMode")
                controller.mainController.shiftToGearIndex(1) -- Park
            end
            
            if batteryLevel >= M.minBatteryPowerLevel then
                forcedChargeMode = false
                electrics.values.chargeMode = 0
                gui.message({txt = "Forced charge-up complete"}, 2, "vehicle.hybrid.chargeMode")
                resumeNormalState()
            else
                if electrics.values.chargeMode ~= 1 and M.powerSwitch then
                    gui.message({txt = "Cannot cancel recovery charge-up. Please wait for battery to charge."}, 5, "vehicle.hybrid.chargeMode")
                    electrics.values.chargeMode = 1
                end
            end
        else
            if gear ~= "P" or electrics.values.evMode == 1 or batteryWorkingRatio >= 0.99 then
                electrics.values.chargeMode = 0
            end

            if electrics.values.chargeMode ~= 1 then
                gui.message({txt = "Charge mode disabled"}, 2, "vehicle.hybrid.chargeMode")
                resumeNormalState()
            end
        end
    elseif state == "stopped" then
        timeSinceEngineCanShutOff = 0
        engine.ignitionCoef = 0
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        electrics.values.iceThrottle = 0
        
        if electrics.values.chargeMode == 1 then
            if gear == "P" then
                engineDemandedOn = true
            else
                gui.message({txt = "Charge mode can only be enabled in P gear"}, 5, "vehicle.hybrid.chargeMode")
                electrics.values.chargeMode = 0
            end
        end

        if electrics.values.evMode == 1 then
            gui.message({txt = "EV mode enabled"}, 2, "vehicle.hybrid.evMode")
            nextState = "evmode"
        elseif gear ~= "N" and engineDemandedOn then
            engine.isDisabled = false -- allow engine to restart

            if electrics.values.rpm < engine.idleRPM then
                nextState = "startup"
                timeSinceEngineStart = 0
                startingTime = 0
            else
                resumeNormalState()
            end
        end
    elseif state == "startup" then
        local targetStarterRPM = engine.idleRPM
        local actualRPM = electrics.values.rpm
        local targetStarterTorque = maprange(actualRPM, targetStarterRPM - 100, targetStarterRPM, M.engineStarterTorque, 0)
        
        engine.ignitionCoef = maprange(actualRPM, engine.idleRPM - 100, engine.idleRPM)
        startingTime = startingTime + dt
        
        if lastState ~= "startup" then
            engine.idleAVStartOffsetSmoother:reset()
            engine.throttleSmoother:reset()
        end
        
        local starterTorque = M.engineStarterSmoother:get(targetStarterTorque, dt)
        
        gearbox:setMG1PowerOverride(maprange(batteryLevel, M.criticalBatteryPowerLevel, M.minBatteryPowerLevel, 0, gearbox.mg1PowerRating))
        gearbox:setMG1TorqueDirection(1)
        
        if gearbox.sunGearAV < (-gearbox.generatorMinFullLoadAV * 5) then
            -- can use backwards generation drag to start engine
            starterTorque = -starterTorque
        end 
        
        gearbox:setRequestedMG1Torque(starterTorque)
        
        if actualRPM >= engine.idleRPM - 200 then
            if engineOn and timeSinceEngineStart > M.engineStartupIdleTime then
                engineStartAttempts = 0
                resumeNormalState()
            end
        else
            M.driveThrottlePid:reset()
            electrics.values.iceThrottle = 0
        end
        
        if startingTime > 3 then
            -- Something's not right; retry at init
            
            timeSinceInit = 0
            engineStartAttempts = engineStartAttempts + 1
            nextState = "init"
        end
    elseif state == "running" then
        engineRunTimeSincePowerOn = engineRunTimeSincePowerOn + dt
        
        if M.maxEngineDeload > 0 then
            engine.ignitionCoef = 1 - M.engineDeloadMaxIgnitionCut * (engineDeload / M.maxEngineDeload)
        else
            engine.ignitionCoef = 1
        end
        
        if not engineOn then
            -- engine died?

            nextState = "startup"
            timeSinceEngineStart = 0
            startingTime = 0
        elseif electrics.values.evMode == 1 then
            if batteryWorkingRatio > 0.026 then
                gui.message({txt = "EV mode enabled"}, 2, "vehicle.hybrid.evMode")
                nextState = "evmode"
            else
                gui.message({txt = "Battery too low; EV mode not available"}, 5, "vehicle.hybrid.evMode")
                electrics.values.evMode = 0
            end
        elseif electrics.values.chargeMode == 1 then
            if gear == "P" then
                gui.message({txt = "Charge mode enabled"}, 2, "vehicle.hybrid.chargeMode")
                chargeThrottle = 0
                nextState = "chargemode"
            else
                gui.message({txt = "Charge mode can only be enabled in P gear"}, 5, "vehicle.hybrid.chargeMode")
                electrics.values.chargeMode = 0
            end
        else
            if engineDemandedOn then
                timeSinceEngineCanShutOff = 0
            else
                if gear ~= "N" then -- engine keeps state in N
                    timeSinceEngineCanShutOff = timeSinceEngineCanShutOff + dt

                    if timeSinceEngineCanShutOff > 1 then
                        stopEngine()
                    end
                end
            end
        end
    elseif state == "poweroff" then
        if M.powerOn then
            M.powerOn = false
            sounds.playSoundOnceFollowNode(M.shutdownSampleName, gearbox.transmissionNodeID, M.shutdownSoundBaseVolume, 1)
        end
        
        if M.powerSwitch then
            if gear == "P" then
                timeSinceInit = 0
                nextState = "init"
                sounds.playSoundOnceFollowNode(M.startupSampleName, gearbox.transmissionNodeID, M.startupSoundBaseVolume, 1)
            else
                M.powerSwitch = false
                gui.message({txt = "Shift to Park before starting vehicle"}, 2, "vehicle.hybrid.starting")
            end
        else
            gearbox:setMG1PowerOverride(0)
            gearbox:setMG2PowerOverride(0)
            gearbox:setInputDisengaged(true)
            timeSinceEngineCanShutOff = 0
            engine:deactivateStarter()
            engine.ignitionCoef = 0
            engine.isDisabled = true
            electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
            electrics.values.iceThrottle = 0
            engineStartAttempts = 0
            forcedChargeMode = false
            shownDisabledMessage = false
            gearLockedInPark = false
        end
    end
    
    lastState = state
end

local function updateBattery(dt)
    local mg1Kw = gearbox.motorGenerator1Power
    local mg2Kw = gearbox.motorGenerator2Power
    
    -- If negative (generating power), multiply by generator efficiency (power loss from motors to battery)
    -- If positive (consumint power), divide by generator efficiency (power loss from battery to motors)
    if mg1Kw < 0 then
        mg1Kw = mg1Kw * M.regenEfficiency
    else
        mg1Kw = mg1Kw / M.regenEfficiency
    end
    
    if mg2Kw < 0 then
        mg2Kw = mg2Kw * M.regenEfficiency
    else
        mg2Kw = mg2Kw / M.regenEfficiency
    end
    
    local idlePowerDraw = M.powerOn and M.idlePowerDraw or 0.0
    
    totalPowerOutKw = idlePowerDraw + mg1Kw + mg2Kw
    
    if M.turboRegen.isExisting then
        totalPowerOutKw = totalPowerOutKw + M.turboRegen.power
    end

    local totalPowerOutKj = totalPowerOutKw * dt

    debugutil.putValue("Battery power (kW)", totalPowerOutKw)
    debugutil.putValue("Generator power (kW)", mg1Kw)

    battery.storedEnergy = mathutil.clamp(battery.storedEnergy - totalPowerOutKj * 1000, 0, battery.initialStoredEnergy)
end

local function updateSounds(dt)
    -- Update sounds
    local mg1Power = abs(gearbox.motorGenerator1Power)
    local mg1AV = abs(gearbox.motorGenerator1AV)
    local mg1InverterPitch = maprange(mg1AV, M.inverterPitchMinSpeedMG1, M.inverterPitchMaxSpeedMG1, M.inverterMinPitchMG1, M.inverterMaxPitchMG1)
    local mg1InverterEffort = min(1, mg1Power / M.inverterMaxVolumeKwMG1)
    local mg1InverterVolume = max(maprange(mg1InverterEffort, 0, 1e-2) * M.inverterMinVolumeMG1, M.inverterBaseVolumeMG1 * maprange(mg1InverterEffort, 1e-2, 1.0))
    
    local mg2Power = abs(gearbox.motorGenerator2Power)
    local mg2AV = abs(gearbox.motorGenerator2AV)
    local mg2InverterPitch = maprange(mg2AV, M.inverterPitchMinSpeedMG2, M.inverterPitchMaxSpeedMG2, M.inverterMinPitchMG2, M.inverterMaxPitchMG2)
    local mg2InverterEffort = min(1, mg2Power / M.inverterMaxVolumeKwMG2)
    local mg2InverterVolume = max(maprange(mg2InverterEffort, 0, 1e-2) * M.inverterMinVolumeMG2, M.inverterBaseVolumeMG2 * maprange(mg2InverterEffort, 1e-2, 1.0))

    obj:setVolumePitch(soundSources.mg1Whine, mg1InverterVolume, M.mg1PitchSmoother:get(mg1InverterPitch, dt))
    obj:setVolumePitch(soundSources.mg2Whine, mg2InverterVolume, M.mg2PitchSmoother:get(mg2InverterPitch, dt))

    local motorPitch = maprange(speedMph, M.motorPitchMinSpeed, M.motorPitchMaxSpeed, M.motorMinPitch, M.motorMaxPitch, false)
    local motorVolume = M.motorInverterEffortVolumeCoef * mg2InverterEffort + maprange(speedMph, M.motorVolumeMinSpeed, M.motorVolumeMaxSpeed, 0, M.motorBaseVolume)

    obj:setVolumePitch(soundSources.motorRamp, motorVolume, motorPitch)
end

local function updateGUI(dt)
    local maxKw
    
    if battery and battery.jbeamData and (battery.jbeamData.maxPowerOutput or battery.jbeamData.maxPowerInput) then
        local maxOutput = battery.jbeamData.maxPowerOutput or 0
        local maxInput = battery.jbeamData.maxPowerInput or 0
        
        maxKw = max(maxOutput, maxInput)
    else
        maxKw = gearbox and max(gearbox.mg1PowerRating, gearbox.mg2PowerRating) or 0
    end
    
    if streams.willSend("hybrid") then
        local str = {
            soc = batteryWorkingRatio,
            kw = totalPowerOutKw,
            maxKw = maxKw,
            minSoc = 0.0,
            maxSoc = 1.0,
        }

        gui.send("hybrid", str)
    end
    
    debugutil.sendStreams()
end

local function updateGFX(dt)
    updateValues(dt)

    local iceThrottle = 0.0
    
    if gearbox and engine and battery then
        calculateElectricLimits(dt)
        calculateTorqueStatistics()
        
        -- Once we know we have all the required components for running,
        -- we can start to perform our (very complex and abundant) hybrid
        -- calculations
        
        -- Calculate some rudimentary locals
        local actualSpeedSign = mathutil.sign(electrics.values.avgWheelAV)
        local desiredSpeedSign = gear == "R" and -1 or 1
        
        -- Pull in some measurements on the gearbox
        local gearboxMaxMG2Torque = gearbox:getMaxMG2Torque()
        local gearboxMaxMG1TorqueUnchecked = gearbox:getMaxMG1Torque(true)
        local gearboxMaxMG2TorqueUnchecked = gearbox:getMaxMG2Torque(true)
        
        -- If there is a maximum power input on the battery, we can't exceed it while regenerating
        -- Instead, we have to figure out the maximum regen coefficient that we can apply without
        -- exceeding the battery's max power input
        local maxRegenTorque = (disableRegen or speedMph < M.minRegenFadeSpeed) and 0 or min(M.maxRegenTorque, gearboxMaxMG2TorqueUnchecked * gearbox.finalMG2GearRatio)
        local maxRegenTorqueOnWheels = maxRegenTorque * gearbox.cumulativeGearRatio
        local uncappedMaxRegenTorque = maxRegenTorque
        
        if battery.jbeamData.maxPowerInput then
            local batteryPowerInputCoef = maprange(batteryLevel, 0.9, 0.95, 1, 0)
            local batteryPowerInputAvailable = battery.jbeamData.maxPowerInput * batteryPowerInputCoef - powerGeneratedFromEngine
            
            -- Calculate what percentage of regen we can use, taking into account
            -- the maximum amount of power the battery can currently accept, minus
            -- however much power the engine is putting into the battery
            maxRegenTorque = min(maxRegenTorque, max(0, mathutil.kwToTorque(batteryPowerInputAvailable, gearbox.motorGenerator2AV) * gearbox.finalMG2GearRatio))
        end
        
        local isInDriveGear = gear ~= "P" and gear ~= "N"
        local currentRegenTorque = speedMph < M.minRegenFadeSpeed and 0 or max(0, gearbox.motorGenerator2Torque * -actualSpeedSign) * gearbox.finalMG2GearRatio
        local currentRegenTorqueOnWheels = currentRegenTorque * gearbox.cumulativeGearRatio
        
        -- Calculate some throttle input variables
        local throttleOffset
        
        if gear == "L" then
            throttleOffset = maprange(maxRegenTorque, 0, uncappedMaxRegenTorque, 0.0001, 0.25)
        else
            throttleOffset = M.evCoastRegen
        end
        
        local posThrottle = math.pow(maprange(throttle, throttleOffset, 1), 1 / M.throttleCurveExponent)
        local negThrottle = maprange(throttle, throttleOffset, 0)
        local requestedThrottle = posThrottle
        local creepThrottle = 0
        local requestedRegenTorque = 0

        -- Fill in some debug values
        debugutil.putValue("Battery SOC % (usable)", math.floor(getProportionalBatteryRatio() * 100))
        debugutil.putValue("Throttle % EV thresh.", math.floor(throttle * 100 / M.maxEvOnlyThrottle))
        
        if brakeInput < 0.25 and electrics.values.parkingbrake < 1 and gear ~= "P" and gear ~= "N" then
            -- Apply creep throttle
            
            creepThrottle = maprange(speedMph, 0, 5, M.evCreepThrottle, 0)
        end
        
        -- Apply hill assist
        local hillAssistAngleCoef = maprange(obj:getDirectionVector().z * desiredSpeedSign, M.minHillAssistPitch, M.maxHillAssistPitch)
        local hillAssistSpeedCoef = maprange(speedMph * actualSpeedSign * desiredSpeedSign, 1e-10, 0.05, 1, 0)
        local hillAssistTorqueCoef = maprange(abs(gearbox.outputTorque1), 0, torqueData.maxCombinedTorque, 1, 0)
        local hillAssistBrakeHold = hillAssistAngleCoef * hillAssistSpeedCoef * hillAssistTorqueCoef * M.hillAssistBrakeHoldAmount
        
        electrics.values.brake = max(electrics.values.brake, hillAssistBrakeHold)

        if brakeInput > 0 then
            -- Updating while brake pedal is applied
            
            if not isInDriveGear or gear == "R" then
                -- No regen in P, R, or N
                
                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Reset brake split coef
                        wheel.brakeInputSplit = initialWheelBrakeSplit[i][1]
                        wheel.brakeSplitCoef = initialWheelBrakeSplit[i][2]
                    end
                end
                
                electrics.values.effectiveBrake = electrics.values.brake

                debugutil.putValue("Friction % of full brakes", math.floor(brakeInput * 100))
            else
                -- Blend friction/regen brakes in any other gear
                
                local wheelsSlipping = false
                local preAdjustedFrictionTorque = 0
                local actualFrictionTorque = 0
                local averageWheelSpeed = 0

                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        preAdjustedFrictionTorque = preAdjustedFrictionTorque + (brakeInput * wheel.brakeTorque)
                        actualFrictionTorque = actualFrictionTorque + wheel.desiredBrakingTorque
                        wheel.totalTorqueCoef = wheel.brakeTorque / totalPossibleBrakeTorque
                        
                        if wheel.isPropulsed then
                            wheel.propulsedTorqueCoef = wheel.brakeTorque / totalPropulsedBrakeTorque
                        else
                            wheel.unpropulsedTorqueCoef = wheel.brakeTorque / totalUnpropulsedBrakeTorque
                        end
                        
                        averageWheelSpeed = averageWheelSpeed + (wheel.wheelSpeed / wheels.wheelCount)
                        
                        if abs(wheel.wheelSpeed - lastAverageWheelSpeed) / lastAverageWheelSpeed > M.maxWheelSlipCoefForRegen then
                            wheelsSlipping = true
                        end
                    end
                end
                
                lastAverageWheelSpeed = averageWheelSpeed

                if disableRegen then
                    if not wheelsSlipping then
                        timeSinceWheelSlip = timeSinceWheelSlip + dt

                        if timeSinceWheelSlip > M.regenWheelslipReenableTime then
                            disableRegen = false
                            timeSinceWheelSlip = 0
                        end
                    end
                else
                    if wheelsSlipping then
                        timeSinceWheelSlip = timeSinceWheelSlip + dt

                        if timeSinceWheelSlip > M.regenWheelslipDisableTime then
                            disableRegen = true
                            timeSinceWheelSlip = 0
                        end
                    else
                        timeSinceWheelSlip = 0
                    end
                end

                -- Combine regen input from brake pedal and throttle pedal (one-pedal driving)
                local regenPercentOfBrakingForce = maxRegenTorqueOnWheels / totalPossibleBrakeTorque
                local regenBrakePedal = maprange(brakeInput, 0, regenPercentOfBrakingForce, 0, 1)
                local regenAccelPedal = gear == "L" and negThrottle or 0
                
                desiredRegen = clamp(regenBrakePedal + regenAccelPedal, 0, 1) * maprange(speedMph, 0, M.minRegenFadeSpeed)
                requestedRegenTorque = desiredRegen * maxRegenTorque

                -- Calculate amount of friction brakes needed to cover brake demand
                local requestedBrakingTorqueOnWheels = brakeInput * totalPossibleBrakeTorque
                local regenBrakePedalCoef = max(0, (desiredRegen - regenAccelPedal) / desiredRegen)
                local currentRegenTorqueOnWheelsFromBrakePedal = regenBrakePedalCoef * currentRegenTorqueOnWheels
                local regenToTotalTorqueRatio = currentRegenTorqueOnWheelsFromBrakePedal / requestedBrakingTorqueOnWheels
                local neededFrictionBrakingTorque = max(0, requestedBrakingTorqueOnWheels - currentRegenTorqueOnWheelsFromBrakePedal)
                local neededFrictionBrakingCoef = clamp(neededFrictionBrakingTorque / totalPossibleBrakeTorque, 0, 1)
                
                currentBrakingCoef = (actualFrictionTorque + currentRegenTorqueOnWheels) / totalPossibleBrakeTorque

                local adjustedFrictionTorque = 0
                
                -- Apply more friction brakes on wheels without regen to balance them
                for i = 0, wheels.wheelCount - 1, 1 do
                    local wd = wheels.wheels[i]
                    local thisWheelCurrentTorque = brakeInput * wd.brakeTorque
                    local thisWheelNeededTorque
                    
                    wd.brakeInputSplit = 0
                    
                    if thisWheelCurrentTorque < 1e-30 then
                        wd.brakeSplitCoef = 1
                    else
                        if wd.isPropulsed then
                            local thisWheelRegenAdjust = min(currentRegenTorqueOnWheels, totalUnpropulsedBrakeTorque) * wd.propulsedTorqueCoef
                            
                            thisWheelNeededTorque = max(0, neededFrictionBrakingTorque - thisWheelRegenAdjust) * wd.totalTorqueCoef
                        else
                            -- Unpropulsed-only braking until friction brakes = regen brakes
                            local thisWheelRegenBalance = min(currentRegenTorqueOnWheels, neededFrictionBrakingTorque) * wd.unpropulsedTorqueCoef
                            -- Then balanced normally amongst propulsed and unpropulsed
                            local thisWheelNormalBalance = max(0, neededFrictionBrakingTorque - currentRegenTorqueOnWheels) * wd.totalTorqueCoef
                            
                            thisWheelNeededTorque = thisWheelRegenBalance + thisWheelNormalBalance
                        end
                        
                        wd.brakeSplitCoef = clamp(thisWheelNeededTorque / thisWheelCurrentTorque, 0, 1)
                    end
                    
                    adjustedFrictionTorque = adjustedFrictionTorque + brakeInput * wd.brakeSplitCoef * wd.brakeTorque
                end
                
                -- Adjust brake input ratio since we may have taken away some brake effort above
                
                electrics.values.brakelights = max(electrics.values.brakelights, 1)
                electrics.values.effectiveBrake = actualFrictionTorque / totalPossibleBrakeTorque

                debugutil.putValue("Actual Regen Torque (Out, Wheels)", fnum(currentRegenTorque) .. ", " .. fnum(currentRegenTorqueOnWheels))
                debugutil.putValue("Requested Braking Torque", requestedBrakingTorqueOnWheels)
                debugutil.putValue("Friction Braking Torque (Req, Act)", fnum(neededFrictionBrakingTorque) .. ", " .. fnum(actualFrictionTorque))
                debugutil.putValue("Total Braking Torque (%)", fnum(currentRegenTorqueOnWheels + preAdjustedFrictionTorque) .. " (" .. fnum(currentBrakingCoef) .. ")")
                debugutil.putValue("Regen % of brake demand", fnum(regenToTotalTorqueRatio * 100, 0))
                debugutil.putValue("Friction % of full brakes", fnum(neededFrictionBrakingCoef * 100, 0))
                debugutil.putValue("Adjusted friction torque", fnum(adjustedFrictionTorque))
                
                --[[ if streams.willSend("profilingData") then
                    gui.send("profilingData", {
                        -- Blue
                        totalBrakeTorque = { title = "Total Brake Torque", color = "#32eaf6", unit = "Nm", value = actualFrictionTorque + currentRegenTorqueOnWheels },
                        -- Purple
                        requestedBrakeTorque = { title = "Requested Brake Torque", color = "#7700ff", unit = "Nm", value = requestedBrakingTorqueOnWheels },
                        -- Green
                        regenBrakeTorque = { title = "Regen Brake Torque", color = "#379637", unit = "Nm", value = currentRegenTorqueOnWheels },
                        -- Red
                        frictionBrakeTorque = { title = "Friction Brake Torque", color = "#f65232", unit = "Nm", value = actualFrictionTorque },
                    })
                end ]]
            end
        else
            disableRegen = false
            currentBrakingCoef = currentRegenTorqueOnWheels / totalPossibleBrakeTorque
            electrics.values.effectiveBrake = electrics.values.brake

            if isInDriveGear then
                -- Regenerative braking when letting off accelerator (tiny bit in D, full in L)
                local zeroThrottleRegen = gear == "L" and 1 or M.evCoastRegen
                
                desiredRegen = negThrottle * zeroThrottleRegen * maprange(speedMph, 0, M.minRegenFadeSpeed)
                requestedRegenTorque = desiredRegen * maxRegenTorque
            else
                desiredRegen = 0
                requestedRegenTorque = 0
            end
            
            debugutil.putValue("Total Braking Torque (%)", fnum(currentRegenTorqueOnWheels) .. " (" .. fnum(currentBrakingCoef) .. ")")
             
            --[[ if streams.willSend("profilingData") then
                gui.send("profilingData", {
                    -- Blue
                    totalBrakeTorque = { title = "Total Brake Torque", color = "#32eaf6", unit = "Nm", value = actualFrictionTorque + currentRegenTorqueOnWheels },
                    -- Purple
                    requestedBrakeTorque = { title = "Requested Brake Torque", color = "#7700ff", unit = "Nm", value = requestedBrakingTorqueOnWheels },
                    -- Green
                    regenBrakeTorque = { title = "Regen Brake Torque", color = "#379637", unit = "Nm", value = currentRegenTorqueOnWheels },
                    -- Red
                    frictionBrakeTorque = { title = "Friction Brake Torque", color = "#f65232", unit = "Nm", value = actualFrictionTorque },
                })
            end ]]
        end
        
        local mg1Dir = 1
        local mg1Torque = 0.0
        local mg2Torque = 0.0
        
        if state == "init" then
            iceThrottle = 0
        elseif gear == "N" then
            iceThrottle = throttle
        elseif state == "chargemode" then
            iceThrottle = chargeThrottle
            
            local chargeTargetRpm = M.fastChargeRpm or M.engineMaxRpm
            local desiredGenPowerCoef = maprange(electrics.values.rpm, chargeTargetRpm - M.engineLoadRpmTolerance, chargeTargetRpm)
            
            mg1Torque = -desiredGenPowerCoef * gearboxMaxMG1TorqueUnchecked
        else
            -- TOO MANY CALCULATIONS, SEND HELP
            
            -- Misc torque variables
            local availablePriorityEvOutputTorque = 0
            local availableSecondaryEvOutputTorque = 0
            local engineTorqueForChargeTorque = 0
            local engineLoad = 0
            
            -- Throttle-related variables
            local normalizedThrottle = requestedThrottle
            local requestedTorque = 0
            local requestedElectricOutputTorque = 0
            local requestedEngineOutputTorque = 0
            
            -- Prevent gearbox or vehicle from going overspeed
            local mg1OverspeedCoef = 0
            local mg2OverspeedCoef = 0
            local vehicleOverspeedCoef = 0
            local maxVehicleSpeed = gear == "R" and M.maxVehicleSpeedReverse or M.maxVehicleSpeed
            local maxVehicleSpeedFalloff = gear == "R" and M.maxVehicleSpeedFalloffReverse or M.maxVehicleSpeedFalloff
            
            -- Other misc upvalues
            local isParked = speedMph < 0.5 and (gear == "P" or electrics.values.parkingbrake == 1)
            local isStopped = speedMph < 0.5 and brakeInput > 0.1
            local launchControl = (brakeInput >= 0.25 or electrics.values.parkingbrake == 1) and throttle > 0.02
            
            -- Limit vehicle speed to spec (with falloff)
            if type(maxVehicleSpeed) == "number" then
                vehicleOverspeedCoef = maprange(speedMph, maxVehicleSpeed, maxVehicleSpeed + maxVehicleSpeedFalloff)
            end
            
            -- Prevent generator motor (MG1) from negative overspeeding ONLY in evmode (engine RPM will prevent this when it's on)
            if state == "evmode" and gearbox.motorGenerator1AV <= 0 then
                mg1OverspeedCoef = maprange(gearbox.motorGenerator1AV, gearbox.minMG1AV + 100, gearbox.minMG1AV)
            end
            
            -- Prevent drive motor (MG2) from overspeeding in either direction
            if gearbox.motorGenerator2AV >= 0 then
                mg2OverspeedCoef = maprange(gearbox.motorGenerator2AV, gearbox.maxMG2AV - 100, gearbox.maxMG2AV)
            else
                mg2OverspeedCoef = maprange(gearbox.motorGenerator2AV, gearbox.minMG2AV + 100, gearbox.minMG2AV)
            end
            
            local overspeedCoef = M.throttleLimiterSmoother:get(mathutil.maxall(
                mg1OverspeedCoef,
                mg2OverspeedCoef,
                vehicleOverspeedCoef
            ), dt)
            
            -- Apply overspeed cap
            normalizedThrottle = min(normalizedThrottle, 1 - overspeedCoef)
            
            -- Calculate RPM range for engine based on required sun gear speeds
            local minRpm = max(M.engineMinIdleRpm, (gearbox:getMinCarrierAV() * gearbox.inputGearRatio) * constants.avToRPM)
            local maxRpm = min(M.engineMaxRpm, (gearbox:getMaxCarrierAV() * gearbox.inputGearRatio) * constants.avToRPM)
            local actualRpm = electrics.values.rpm
            
            -- Figure out max possible engine torque at its current RPM, as well as the torque it's currently
            -- applying on\the different transmission components
            local achievableEngineTorque = engineutil.getAchievableEngineTorque(engine)
            local achievableEngineTorqueCarrier = achievableEngineTorque * gearbox.inputGearRatio
            local achievableEngineTorqueSun = abs(gearbox:getSunTorqueFromCarrierTorque(achievableEngineTorqueCarrier))
            local achievableEngineTorqueRing = abs(gearbox:getRingTorqueFromCarrierTorque(achievableEngineTorqueCarrier))
            local actualEngineTorque = max(0, engine.outputTorque1)
            local actualEngineTorqueCarrier = actualEngineTorque * gearbox.inputGearRatio
            local actualEngineTorqueRing = abs(gearbox:getRingTorqueFromCarrierTorque(actualEngineTorqueCarrier))
            local actualEngineTorqueSun = abs(gearbox:getSunTorqueFromCarrierTorque(actualEngineTorqueCarrier))
            
            maxRpm = max(minRpm, maxRpm)
            
            if state == "running" then
                -- Calculate desired charge power/torque
                
                local chargePowerCoef = 0
                
                if isParked or isStopped then
                    -- Charge while stopped. Charge power is achieved using pure MG1 drag on ICE.
                    
                    if isInChargeup then
                        requestedChargePower = M.chargePowerSmoother:get(M.idlePowerDraw + max(M.chargeupMinPower, isParked and M.maxChargePowerParked or 0.0), dt)
                        chargePowerCoef = 1.0
                    else
                        if isParked then
                            chargePowerCoef = 1.0
                        else
                            chargePowerCoef = 0.25
                        end
                        
                        requestedChargePower = M.chargePowerSmoother:get(M.idlePowerDraw + M.maxChargePowerParked * chargePowerCoef * maprange(batteryWorkingRatio, 0.5, max(0, M.batteryChargeLowRatio)), dt)
                        
                        if isParked and type(M.parkedChargingMinIdleRpm) == "number" then
                            local parkedChargingCoef = maprange(requestedChargePower, 0, M.maxChargePowerParked)
                            
                            minRpm = max(minRpm, maprange(parkedChargingCoef, 0, 1, minRpm, M.parkedChargingMinIdleRpm))
                        end
                    end
                    
                    local requestedChargeTorque = mathutil.kwToTorque(requestedChargePower / M.regenEfficiency, gearbox.motorGenerator1AV)
                    
                    engineTorqueForChargeTorque = abs(gearbox:getCarrierTorqueFromSunTorque(requestedChargeTorque))
                else
                    local chargePowerThrottleCoef = M.maxChargePowerCoefThrottle:get(normalizedThrottle * 100)
                    local chargePowerBatteryCoef = maprange(batteryWorkingRatio, M.targetRelativeBatteryLevel, M.targetRelativeBatteryLevel - M.targetBatteryLevelTolerance)
                    local chargePowerRegenCoef = maprange(desiredRegen, 0, 1, 1, 0)
                    local chargePowerFactor = gear == "R" and 0 or 1
                    local maxChargePower = M.maxChargePowerVsSpeed:get(speedMph)
                    
                    chargePowerCoef = chargePowerThrottleCoef * chargePowerBatteryCoef * chargePowerRegenCoef
                    requestedChargePower = M.chargePowerSmoother:get(min(chargePowerCoef, chargePowerFactor) * maxChargePower, dt)
                end
            else
                requestedChargePower = 0
            end
            
            if isInDriveGear then
                -- Update while in a driving gear (anything other than Park or Neutral)
                
                -- Map throttle to total capable torque output
                if type(M.evOnlyThrottleDetent) == "number" then
                    -- There's a specific "detent" in the accelerator where the engine will kick in, if the battery is full
                    -- This makes it easier to keep the vehicle in EV mode with a full battery
                    
                    local requestedEvTorque = maprange(normalizedThrottle, 0, M.evOnlyThrottleDetent) * torqueData.fullBatteryElectricTorque
                    local requestedEngineTorque = maprange(normalizedThrottle, M.evOnlyThrottleDetent, 1.001) * achievableEngineTorque
                    
                    requestedTorque = min(torqueData.maxCombinedTorque, requestedEvTorque + requestedEngineTorque)
                else
                    requestedTorque = normalizedThrottle * torqueData.maxCombinedTorque
                end
                
                -- Calculate available EV torque and distribute torque demand between motors and engine
                if state == "evmode" then
                    local gearboxMaxMG1Torque = gearbox:getMaxMG1Torque(availablePriorityEvPower)
                    local availableMG1OutputTorque = gearbox:getRingTorqueFromSunTorque(gearboxMaxMG1Torque) * gearbox.finalRingGearRatio
                    local availableMG2OutputTorque = gearbox:getMaxMG2Torque(availablePriorityEvPower) * gearbox.finalMG2GearRatio
                    
                    availablePriorityEvOutputTorque = availableMG1OutputTorque + availableMG2OutputTorque
                    availableSecondaryEvOutputTorque = 0
                else
                    availablePriorityEvOutputTorque = gearbox:getMaxMG2Torque(availablePriorityEvPower) * gearbox.finalMG2GearRatio
                    availableSecondaryEvOutputTorque = gearbox:getMaxMG2Torque(availableSecondaryEvPower) * gearbox.finalMG2GearRatio
                end
                
                local chargeOutputTorque = speedMph < 0.5 and 0 or mathutil.kwToTorque(requestedChargePower / M.regenEfficiency, gearbox.motorGenerator2AV) * gearbox.finalMG2GearRatio
                local torqueStillNeeded = requestedTorque
                local creepTorque = creepThrottle * torqueData.maxElectricTorque
                
                -- Consume available primary EV power first (from whatever is available from battery)
                if ecvtDriveMode == 0 then
                    -- Apply charge torque here
                    requestedElectricOutputTorque = min(torqueStillNeeded, availablePriorityEvOutputTorque - chargeOutputTorque) + creepTorque
                else
                    -- Apply charge torque later
                    requestedElectricOutputTorque = min(torqueStillNeeded, availablePriorityEvOutputTorque) + creepTorque
                end
                
                -- Have to adjust for creepTorque here so as not to confuse the following calculations
                torqueStillNeeded = torqueStillNeeded + creepTorque - requestedElectricOutputTorque
                
                -- If we would use too much secondary EV power, start the ICE (or keep it on)
                local secondaryPowerDemand = mathutil.torqueToKw(torqueStillNeeded, gearbox.motorGenerator2AV)
                    
                if secondaryPowerDemand > M.maxEvOnlySecondaryPower then
                    requestEngine("secondary power exceeded")
                end
                
                if state == "running" then
                    -- Try and distribute torque demand to engine in "balanced mode" where the power consumption of MG2
                    -- equals the power generated by MG1 (with a slight adjustment towards draining or charging depending
                    -- on the state of the battery)
                    
                    -- Very length algebraic equation to figure out the neutral-balance of engine and electric torque
                    local tO = max(0, torqueStillNeeded)
                    local rRC = gearbox.ringCarrierRatio
                    local rSC = gearbox.sunCarrierRatio
                    local rFR = gearbox.finalRingGearRatio
                    local rF2 = gearbox.finalMG2GearRatio
                    local av1 = max(1e-30, abs(gearbox.motorGenerator1AV))
                    local av2 = max(1e-30, abs(gearbox.motorGenerator2AV))
                    
                    -- warning: super ugly equation below
                    requestedEngineOutputTorque = min(torqueData.maxEngineTorque, (tO * rRC * rFR) / (rRC * rFR + ((rSC * av1 * rF2) / av2)))
                end
                
                -- While the engine is working on revving up, the main drive motor should compensate for it
                -- (using secondary power) so that the driver feels a more immediate acceleration response.
                -- It's okay if we do this even if the battery is low (it's a short period of time it will be drawing power)
                -- By using "actualEngineOutputTorque" and subtracting it from torqueStillNeeded, secondary power
                -- will automatically kick in for this if the engine is producing less than expected
                local actualEngineOutputTorque = max(0, actualEngineTorqueRing) * gearbox.finalRingGearRatio
                
                torqueStillNeeded = max(0, torqueStillNeeded - actualEngineOutputTorque)
                
                -- Distribute as much remaining torque demand as we can to secondary EV power
                requestedElectricOutputTorque = requestedElectricOutputTorque + min(torqueStillNeeded, availableSecondaryEvOutputTorque)
                torqueStillNeeded = torqueStillNeeded - min(torqueStillNeeded, availableSecondaryEvOutputTorque)
                
                -- Make sure engine starts if we're out of electric power
                if gear ~= "R" and (torqueStillNeeded > 1e-20 or requestedElectricOutputTorque > torqueData.maxElectricTorque) then
                    requestEngine("electric torque exhausted")
                end
                
                if ecvtDriveMode == 1 then
                    -- Apply charge torque here in drive mode 1
                    torqueStillNeeded = torqueStillNeeded + chargeOutputTorque
                end
                
                -- At this point if torque demand still remains, the battery is probably VERY low. We dump any remaining torque
                -- demand into the engine, which will cause it to rev up quite a bit and rapidly charge the battery at low
                -- speeds.
                if state == "running" and torqueStillNeeded >= 1e-30 then
                    requestedEngineOutputTorque = min(torqueData.maxEngineTorque, requestedEngineOutputTorque + torqueStillNeeded)
                end
                
                if gear ~= "R" and not miscEngineDemand then
                    if throttle > M.maxEvOnlyThrottle then
                        requestEngine("ev throttle limit exceeded")
                    end
                    if launchControl then
                        requestEngine("launch control")
                    end
                end
                
                if brakeInput > 0.1 or electrics.values.parkingbrake > 0 then
                    requestedElectricOutputTorque = min(M.maxLaunchControlTorque, requestedElectricOutputTorque)
                end
                
                debugutil.putValue("Available EV Torque (Priority/Secondary)", fnum(availablePriorityEvOutputTorque) .. " / " .. fnum(availableSecondaryEvOutputTorque))
            else
                requestedTorque = 0
                requestedElectricOutputTorque = 0
                requestedEngineOutputTorque = maprange(normalizedThrottle, M.evCreepThrottle, 1, 0, M.maxParkEngineTorqueCoef * torqueData.maxEngineTorque)
                
                if throttle > 0.02 then
                    requestEngine("throttle in park")
                end
            end
            
            -- apply a final cap to both torques to make sure they don't independently break the max output torque
            requestedElectricOutputTorque = min(torqueData.maxCombinedTorque, requestedElectricOutputTorque)
            requestedEngineOutputTorque = min(torqueData.maxCombinedTorque, requestedEngineOutputTorque)
            
            local requestedEngineTorqueCarrier = engineTorqueForChargeTorque + abs(gearbox:getCarrierTorqueFromRingTorque(requestedEngineOutputTorque / gearbox.finalRingGearRatio))
            local requestedEngineTorqueSun = abs(gearbox:getSunTorqueFromCarrierTorque(requestedEngineTorqueCarrier))
            local requestedEngineTorqueRing = abs(gearbox:getRingTorqueFromCarrierTorque(requestedEngineTorqueCarrier))
            local requestedEngineTorque = requestedEngineTorqueCarrier / gearbox.inputGearRatio
            
            debugutil.putValue("Achievable Engine Torque (Eng, S, C, R)", fnum(achievableEngineTorque, 0) .. ", " ..
                                                                          fnum(achievableEngineTorqueSun, 0) .. ", " ..
                                                                          fnum(achievableEngineTorqueCarrier, 0) .. ", " ..
                                                                          fnum(achievableEngineTorqueRing, 0))
            debugutil.putValue("Requested Engine Torque (Eng, S, C, R)", fnum(requestedEngineTorque, 0) .. ", " ..
                                                                         fnum(requestedEngineTorqueSun, 0) .. ", " ..
                                                                         fnum(requestedEngineTorqueCarrier, 0) .. ", " ..
                                                                         fnum(requestedEngineTorqueRing, 0))
            
            local maxMG1TorqueCarrierAV = gearbox:getCarrierAVFromSunAV(gearbox.mg1PeakEfficiencyAv)
            local maxMG1TorqueRpm = maxMG1TorqueCarrierAV * gearbox.inputGearRatio * constants.avToRPM
            
            if isInDriveGear and not launchControl then
                local minTargetRpm = minRpm
                local maxTargetRpm = maxRpm
                
                targetEngineRpm = clamp(M.engineRpmVsTorque:get(requestedEngineTorque), minTargetRpm, maxTargetRpm)
                
                debugutil.putValue("Min Target RPM", minTargetRpm)
                debugutil.putValue("Max Target RPM", maxTargetRpm)
                debugutil.putValue("MG1 Max Torque Carrier AV", maxMG1TorqueCarrierAV)
                debugutil.putValue("MG1 Max Torque RPM", maxMG1TorqueRpm)
            else
                local minTargetRpm = minRpm
                local maxTargetRpm = maxRpm
                
                targetEngineRpm = maprange(normalizedThrottle, 0, 1, minTargetRpm, maxTargetRpm)
            end
                
            if state == "running" then
                -- Calculate RPM load bias
                local rpmError = maprange(actualRpm, targetEngineRpm - M.engineLoadRpmTolerance, targetEngineRpm + M.engineLoadRpmTolerance, -1, 1)
                local loadBiasCoef = rpmError * (rpmError > 0 and M.engineMaxPositiveLoadBias or M.engineMaxNegativeLoadBias)
                local loadBiasTorque = loadBiasCoef * gearbox:getMaxMG1Torque(true)
                
                engineLoad = max(0, requestedEngineTorqueSun + loadBiasTorque)
                
                -- Calculate engine deloading
                local availableEngineDeloadBattery = maprange(batteryWorkingRatio, 0, 0.1, 0, 1)
                local availableEngineDeloadRegen = maprange(requestedRegenTorque, 0, maxRegenTorque, 0, 1)
                local availableEngineDeload = M.maxEngineDeload * mathutil.maxall(availableEngineDeloadBattery, availableEngineDeloadRegen)
                
                -- Apply engine deloading
                engineDeload = maprange(desiredRegen, 0, 0.05, 0, availableEngineDeload)
                engineLoad = engineLoad - engineDeload
                
                -- Limit engine load based on certain criteria
                local maxEngineLoadStall = maprange(electrics.values.rpm, engine.idleRPM - 200, engine.idleRPM)
                local maxEngineLoadOverspeed = maprange(mathutil.maxall(mg2OverspeedCoef, vehicleOverspeedCoef), 1, 0)
                local maxEngineLoadBattery = ecvtDriveMode == 1 and 1 or maprange(batteryLevel, 1, 0.98)
                
                engineLoad = engineLoad * mathutil.minall(maxEngineLoadStall, maxEngineLoadOverspeed, maxEngineLoadBattery)
            
                debugutil.putValue("RPM Error", rpmError)
                debugutil.putValue("Load Bias", loadBiasTorque)
            else
                engineLoad = 0
                engineDeload = 0
            end
            
            if state == "stopped" or state == "evmode" or not M.powerOn then
                M.driveThrottlePid:reset()
                iceThrottle = 0
            else
                -- Engine throttle is controlled by the desired output torque of the engine
                -- Engine RPM is controlled by the throttle pedal
                -- Actual load on the engine is biased to achieve the desired RPM
                M.driveThrottlePid:setILock(false)
                
                if M.engineControlMethod == "torque" then
                    iceThrottle = M.driveThrottlePid:get(requestedEngineTorque, actualEngineTorque, dt)
                else -- "rpm"
                    iceThrottle = M.driveThrottlePid:get(targetEngineRpm, actualRpm, dt)
                end
            end
            
            -- Calculate actual torques
            local actualTorque = gearbox.outputTorque1
            local actualTorqueEV = gearbox.motorGenerator2Torque * gearbox.finalMG2GearRatio
            local actualTorqueICE = -gearbox:getRingTorqueFromCarrierTorque(actualEngineTorqueCarrier) * gearbox.finalRingGearRatio
            
            if state ~= "running" then
                ecvtDriveMode = 0
                M.latchDriveModeZero:reset()
                M.latchDriveModeOne:reset()
            end
            
            -- Determine actual requested torque from requested regen coef
            local stationaryCarrierSunAV = gearbox:getSunAVFromCarrierAV(0.0)
            local sunAVError = (gearbox.motorGenerator1AV - stationaryCarrierSunAV) / 20
            local engineShutdownLoad = maprange(abs(sunAVError), 0.02, 1, 0, M.engineSpooldownTorque)
            
            -- Determine MG2 adjustment/compensation values
            if state == "startup" and isInDriveGear then
                local starterTorqueRingCompensation = gearbox:getRingTorqueFromSunTorque(gearbox.motorGenerator1Torque)
                local starterTorqueFinalCompensation = starterTorqueRingCompensation * gearbox.finalRingGearRatio
                local starterTorqueMG2Compensation = starterTorqueFinalCompensation / gearbox.finalMG2GearRatio
                
                mg2TorqueAdjustment = starterTorqueMG2Compensation
            end
            
            if state == "evmode" then
                -- Use both motors to achieve the requested torque
                
                local maxMG2TorqueToOutput = gearboxMaxMG2Torque * gearbox.finalMG2GearRatio
                local neededMG2TorqueToOutput = min(requestedElectricOutputTorque, maxMG2TorqueToOutput)
                local neededMG1CompensationToOutput = gear == "R" and 0 or max(0, requestedElectricOutputTorque - maxMG2TorqueToOutput)
                local neededMG2Torque = (abs(neededMG2TorqueToOutput) - requestedRegenTorque) / gearbox.finalMG2GearRatio
                local neededMG1Torque = abs(gearbox:getSunTorqueFromRingTorque(neededMG1CompensationToOutput / gearbox.finalRingGearRatio))
                
                if type(M.maxEvModeMG1Torque) == "number" then
                    neededMG1Torque = min(M.maxEvModeMG1Torque, neededMG1Torque)
                end
                
                mg2Torque = neededMG2Torque
                mg1Torque = neededMG1Torque - max(0, sign(sunAVError) * engineShutdownLoad)
                mg1Dir = -1
            else
                local neededMG2Torque = (requestedElectricOutputTorque - requestedRegenTorque) / gearbox.finalMG2GearRatio
                
                if ecvtDriveMode == 0 then
                    -- engine load can be achieved using gen drag on MG1
                    -- charging is done by adding load to the engine
                    
                    if state == "startup" or state == "running" then
                        mg1Torque = -engineLoad
                        mg1Dir = 1
                    else
                        -- To shut down engine faster, we apply engine load proportional to the engine RPM (but only if MG1 is spinning positive)
                        
                        mg1Torque = engineShutdownLoad
                        mg1Dir = -sign(sunAVError)
                    end
                    
                    mg2Torque = neededMG2Torque + mg2TorqueAdjustment
                else
                    -- MG1 spinning too slow (or backwards) for gen drag; use negative power instead
                    -- in this case, we have to regen off of MG2 to provide power for MG1
                    -- (thus we have to limit MG1 to what MG2 is able to counteract)
                    
                    local engineLoadPower = mathutil.torqueToKw(max(0, engineLoad), abs(gearbox.motorGenerator1AV))
                    local batteryPowerAvailableForEngineLoad = availablePriorityEvPower -- can use priority power to apply engine load
                    local neededPowerForEngineLoad = max(0, engineLoadPower - batteryPowerAvailableForEngineLoad)
                    local neededRegenTorque = abs(mathutil.kwToTorque(neededPowerForEngineLoad + requestedChargePower, gearbox.motorGenerator2AV))
                    
                    mg1Torque = abs(engineLoad)
                    mg1Dir = -sign(engineLoad)
                    mg2Torque = neededMG2Torque - neededRegenTorque + mg2TorqueAdjustment
                end
            end
            
            if state == "evmode" then
                ecvtDriveMode = 0
            else
                -- Determine which drive mode we should be in next based on whether or not MG1 can provide generation drag
                local modeZeroLatchTest = state ~= "running"
                    or gearbox.sunGearAV > gearbox.generatorMinFullLoadAV + 150
                local modeOneLatchTest = state == "running" 
                    and gearbox.sunGearAV < gearbox.generatorMinFullLoadAV * 1.5
                
                if M.latchDriveModeOne:test(modeOneLatchTest, dt) then
                    if ecvtDriveMode == 0 then
                        M.latchDriveModeZero:reset()
                        ecvtDriveMode = 1
                    end
                end
                if M.latchDriveModeZero:test(modeZeroLatchTest, dt) then
                    if ecvtDriveMode == 1 then
                        M.latchDriveModeOne:reset()
                        ecvtDriveMode = 0
                    end
                end
            end
            
            debugutil.putValue("Requested Torque (Total, EV, ICE)", fnum(requestedTorque) .. ", " .. fnum(requestedElectricOutputTorque) .. ", " .. fnum(requestedEngineOutputTorque))
            debugutil.putValue("Actual Torque (Total, EV, ICE)", fnum(actualTorque) .. ", " .. fnum(actualTorqueEV) .. ", " .. fnum(actualTorqueICE))
            debugutil.putValue("Available MG1 Power", fnum(availableMG1Power))
            debugutil.putValue("Available MG2 Power", fnum(availableMG2Power))
            debugutil.putValue("Desired ICE Throttle", iceThrottle)
            debugutil.putValue("Max Regen Torque (Out, Wheels)", fnum(maxRegenTorque) .. ", " .. fnum(maxRegenTorque * gearbox.cumulativeGearRatio))
            debugutil.putValue("Requested Regen Torque (Out, Wheels)", fnum(requestedRegenTorque) .. ", " .. fnum(requestedRegenTorque * gearbox.cumulativeGearRatio))
            debugutil.putValue("Engine Load Torque", engineLoad)
            debugutil.putValue("Engine Deload Torque", engineDeload)
            debugutil.putValue("Min RPM", minRpm)
            debugutil.putValue("Max RPM", maxRpm)
            debugutil.putValue("Target RPM", targetEngineRpm)
            debugutil.putValue("Actual RPM", actualRpm)
            debugutil.putValue("Actual Engine Torque (Eng, S, C, R)", fnum(actualEngineTorque, 0) .. ", " ..
                                                                      fnum(actualEngineTorqueSun, 0) .. ", " ..
                                                                      fnum(actualEngineTorqueCarrier, 0) .. ", " ..
                                                                      fnum(actualEngineTorqueRing, 0))
            debugutil.putValue("eCVT Drive Mode", ecvtDriveMode)
            debugutil.putValue("Available Power (Priority/Secondary)", fnum(availablePriorityEvPower) .. " / " .. fnum(availableSecondaryEvPower))
            debugutil.putValue("Requested Charge Power", fnum(requestedChargePower) .. " kW")
        end
        
        gearbox:setMG1TorqueDirection(mg1Dir)
        gearbox:setRequestedMG1Torque(mg1Torque)
        gearbox:setRequestedMG2Torque(mg2Torque)
        
        -- Calculate motor power restrictions
        local mg2GeneratedPower = max(-gearbox.motorGenerator2Power, 0)
        
        if ecvtDriveMode == 0 then
            availableMG1Power = min(gearbox.mg1PowerRating, availableBatteryPower)
        else
            availableMG1Power = min(gearbox.mg1PowerRating, availableBatteryPower + mg2GeneratedPower)
        end
        
        availableMG2Power = min(gearbox.mg2PowerRating, availableSecondaryEvPower) -- Secondary is always battery + engine generated, so use that
        
        -- Limit the amount of power MG1 and MG2 are allowed to consume (so they don't overdraw the battery)
        gearbox:setMG1PowerOverride(M.mg1MaxPowerSmoother:get(availableMG1Power, dt))
        gearbox:setMG2PowerOverride(M.mg2MaxPowerSmoother:get(availableMG2Power, dt))
        
        updateState(dt)
        updateBattery(dt)
    end
        
    if M.turboRegen.isExisting then
        local maxTurboRegenPowerCoef = M.powerOn and maprange(batteryWorkingRatio, 0, 0.05) or 0
        local maxTurboRegenPower = M.turboRegen.powerRating * maxTurboRegenPowerCoef
        
        M.turboRegen:setMaxPower(maxTurboRegenPower)
        M.turboRegen:updateGFX(dt)
    end
    
    electrics.values.iceThrottle = iceThrottle

    if battery then
        debugutil.putValue("Battery SOC %", tostring(battery.remainingRatio * 100))
    end

    debugutil.putValue("Input Throttle", (throttle or 0))
    debugutil.putValue("Actual ICE Throttle", electrics.values.iceThrottle)
    debugutil.putValue("ICE Load", engine.engineLoad)
    debugutil.putValue("Throttle Aggr.", throttleAggressiveness)

    if nextState then
        state = nextState
    end

    debugutil.putValue("ICE Torque", engine.outputTorque1)
    debugutil.putValue("ICE HP", mathutil.torqueToKw(engine.outputTorque1, engine.outputAV1) * 1.34102)
    debugutil.putValue("Charge Throttle", chargeThrottle)
    debugutil.putValue("MG2 Adjustment", mg2TorqueAdjustment)

    updateGUI(dt)
    updateSounds(dt)
end

local function init(jbeamData)
    M.jbeamData = jbeamData
    
    M.turboRegen = nil
    
    M.startEvMode = jbeamData.startEvMode == true
    M.forceEvMode = jbeamData.forceEvMode == true
    M.startPoweredOff = jbeamData.startPoweredOff or false
    M.powerSwitch = not M.startPoweredOff
    M.powerOn = not M.startPoweredOff
    M.initializeTime = jbeamData.initializeTime or 2.0
    M.vehicleStartupEngineDelay = jbeamData.vehicleStartupEngineDelay or 10
    M.vehicleStartupEngineRuntime = jbeamData.vehicleStartupEngineRuntime or 15
    M.hasAuxiliaryStarterMotor = jbeamData.hasAuxiliaryStarterMotor or false
    
    M.maxEvOnlySecondaryPower = jbeamData.maxEvOnlySecondaryPower or 25 -- kW
    M.maxEvOnlySpeed = jbeamData.maxEvOnlySpeed or 65 -- MPH
    M.maxEvOnlyThrottle = jbeamData.maxEvOnlyThrottle or 0.65 -- ratio
    M.maxEvOnlyThrottleAggressiveness = jbeamData.maxEvOnlyThrottleAggressiveness or 20
    M.maxEvOnlyPitch = jbeamData.maxEvOnlyPitch or math.pi -- 180 degrees pitch by default; will never happen (effectively disabled)
    M.batteryChargeLowRatio = jbeamData.batteryChargeLowRatio or 0.0
    M.batteryChargeHighRatio = jbeamData.batteryChargeHighRatio or 0.15
    
    M.maxEngineAssistSecondaryPower = jbeamData.maxEngineAssistSecondaryPower or 65 -- kW
    
    M.maxEvModeMG1Torque = jbeamData.maxEvModeMG1Torque
    M.evOnlyThrottleDetent = jbeamData.evOnlyThrottleDetent
    M.engineShutOffThrottleThreshold = jbeamData.engineShutOffThrottleThreshold or 0.10
    
    M.minRegenFadeSpeed = jbeamData.minRegenFadeSpeed or 3 -- MPH
    M.maxRegenFadeSpeed = jbeamData.maxRegenFadeSpeed or 8 -- MPH
    
    M.idlePowerDraw = jbeamData.idlePowerDraw or 1 -- kW
    M.minBatteryPowerLevel = jbeamData.minBatteryPowerLevel or 0.25 -- ratio; low end of max-kW slope
    M.maxBatteryPowerLevel = jbeamData.maxBatteryPowerLevel or 0.85 -- ratio; high end of max-kW slope
    M.criticalBatteryPowerLevel = jbeamData.criticalBatteryPowerLevel or 0.05 -- raw state-of-charge
    M.targetRelativeBatteryLevel = jbeamData.targetRelativeBatteryLevel or 0.45 -- 45%
    M.targetBatteryLevelTolerance = jbeamData.targetBatteryLevelTolerance or 0.20 -- 20%
    M.maxFastChargePower = jbeamData.maxFastChargePower or 30 -- kW
    M.fastChargeRpm = jbeamData.fastChargeRpm
    
    M.enableAutoStop = jbeamData.enableAutoStop ~= false
    M.autoStopBrakeThreshold = jbeamData.autoStopBrakeThreshold or 0
    M.autoStopBrakeTime = jbeamData.autoStopBrakeTime or 0
    M.engineMinRunTime = jbeamData.engineMinRunTime or 4 -- seconds
    M.minCoolantTemp = jbeamData.minCoolantTemp or 30
    M.minCoolantTempWarmedUp = jbeamData.minCoolantTempWarmedUp or 57
    
    M.engineMaxRpm = jbeamData.engineMaxRpm
    M.engineStartupIdleTime = jbeamData.engineStartupIdleTime or 0.15
    M.engineStarterTorque = jbeamData.engineStarterTorque or 28
    M.engineSpooldownTorque = jbeamData.engineSpooldownTorque or 15
    M.engineControlMethod = jbeamData.engineControlMethod or "rpm"
    M.maxChargePowerParked = jbeamData.maxChargePowerParked or 3
    M.parkedChargingMinIdleRpm = jbeamData.parkedChargingMinIdleRpm
    M.chargeupMinPower = jbeamData.chargeupMinPower or 3
    M.engineMinIdleRpm = jbeamData.engineMinIdleRpm
    M.engineLoadRpmTolerance = jbeamData.engineLoadRpmTolerance or 550
    M.engineMaxPositiveLoadBias = jbeamData.engineMaxPositiveLoadBias or 0.25
    M.engineMaxNegativeLoadBias = jbeamData.engineMaxNegativeLoadBias or 0.25
    M.maxEngineDeload = jbeamData.maxEngineDeload or 5
    M.engineDeloadMaxIgnitionCut = jbeamData.engineDeloadMaxIgnitionCut or 0.4
    M.maxParkEngineTorqueCoef = jbeamData.maxParkEngineTorqueCoef or 0.5
    M.maxVehicleSpeed = jbeamData.maxVehicleSpeed
    M.maxVehicleSpeedReverse = jbeamData.maxVehicleSpeedReverse or 20
    M.maxVehicleSpeedFalloff = jbeamData.maxVehicleSpeedFalloff or 5
    M.maxVehicleSpeedFalloffReverse = jbeamData.maxVehicleSpeedFalloffReverse or M.maxVehicleSpeedFalloff
    M.maxLaunchControlTorque = jbeamData.maxLaunchControlTorque or 15
    
    M.minThrottleAggressivenessRamp = jbeamData.minThrottleAggressivenessRamp or 1.5
    M.maxThrottleAggressivenessRamp = jbeamData.maxThrottleAggressivenessRamp or 8
    M.minThrottleAggressivenessFactor = jbeamData.minThrottleAggressivenessFactor or 1
    M.maxThrottleAggressivenessFactor = jbeamData.maxThrottleAggressivenessFactor or 10
    M.throttleCurveExponent = jbeamData.throttleCurveExponent or 1.1
    
    M.maxRegenTorque = jbeamData.maxRegenTorque or 200
    M.maxOutputTorque = jbeamData.maxOutputTorque
    M.maxTorqueBoostPerDegree = jbeamData.maxTorqueBoostPerDegree or 0
    M.maxTorqueBoost = jbeamData.maxTorqueBoost or 0
    M.regenEfficiency = jbeamData.regenEfficiency or 0.93
    M.regenWheelslipDisableTime = jbeamData.regenWheelslipDisableTime or 0.5
    M.regenWheelslipReenableTime = jbeamData.regenWheelslipReenableTime or 1.0
    M.maxWheelSlipCoefForRegen = jbeamData.maxWheelSlipCoefForRegen or 0.05
    
    M.evCreepThrottle = jbeamData.evCreepThrottle or 0.25
    M.evCoastRegen = jbeamData.evCoastRegen or 0.05
    M.hillAssistBrakeHoldAmount = jbeamData.hillAssistBrakeHoldAmount or 0.15 -- %
    M.minHillAssistPitch = jbeamData.minHillAssistPitch or 0.0523599 -- rads; 3 degrees
    M.maxHillAssistPitch = jbeamData.maxHillAssistPitch or 0.20944 -- rads; 12 degrees
    
    M.inverterSampleNameMG1 = jbeamData.inverterSampleNameMG1 or jbeamData.inverterSampleName or "vehicles/hybrid_common/sounds/inverter.ogg"
    M.inverterMaxVolumeKwMG1 = jbeamData.inverterMaxVolumeKwMG1 or jbeamData.inverterMaxVolumeKw or 20
    M.inverterBaseVolumeMG1 = jbeamData.inverterBaseVolumeMG1 or jbeamData.inverterBaseVolume or 0.35
    M.inverterMinVolumeMG1 = jbeamData.inverterMinVolumeMG1 or jbeamData.inverterMinVolume or 0.05
    M.inverterPitchMinSpeedMG1 = jbeamData.inverterPitchMinSpeedMG1 or jbeamData.inverterPitchMinSpeed or 1
    M.inverterPitchMaxSpeedMG1 = jbeamData.inverterPitchMaxSpeedMG1 or jbeamData.inverterPitchMaxSpeed or 1100
    M.inverterMinPitchMG1 = jbeamData.inverterMinPitchMG1 or jbeamData.inverterMinPitch or 0.05
    M.inverterMaxPitchMG1 = jbeamData.inverterMaxPitchMG1 or jbeamData.inverterMaxPitch or 2.0
    
    M.inverterSampleNameMG2 = jbeamData.inverterSampleNameMG2 or jbeamData.inverterSampleName or "vehicles/hybrid_common/sounds/inverter.ogg"
    M.inverterMaxVolumeKwMG2 = jbeamData.inverterMaxVolumeKwMG2 or jbeamData.inverterMaxVolumeKw or 20
    M.inverterBaseVolumeMG2 = jbeamData.inverterBaseVolumeMG2 or jbeamData.inverterBaseVolume or 0.35
    M.inverterMinVolumeMG2 = jbeamData.inverterMinVolumeMG2 or jbeamData.inverterMinVolume or 0.05
    M.inverterPitchMinSpeedMG2 = jbeamData.inverterPitchMinSpeedMG2 or jbeamData.inverterPitchMinSpeed or 1
    M.inverterPitchMaxSpeedMG2 = jbeamData.inverterPitchMaxSpeedMG2 or jbeamData.inverterPitchMaxSpeed or 1100
    M.inverterMinPitchMG2 = jbeamData.inverterMinPitchMG2 or jbeamData.inverterMinPitch or 0.05
    M.inverterMaxPitchMG2 = jbeamData.inverterMaxPitchMG2 or jbeamData.inverterMaxPitch or 2.0
    
    M.motorSampleName = jbeamData.motorSampleName or "vehicles/hybrid_common/sounds/motor_ramp.ogg"
    M.motorBaseVolume = jbeamData.motorBaseVolume or 0
    M.motorVolumeMinSpeed = jbeamData.motorVolumeMinSpeed or 4
    M.motorVolumeMaxSpeed = jbeamData.motorVolumeMaxSpeed or 15
    M.motorPitchMinSpeed = jbeamData.motorPitchMinSpeed or 0
    M.motorPitchMaxSpeed = jbeamData.motorPitchMaxSpeed or 50
    M.motorMinPitch = jbeamData.motorMinPitch or 0.01
    M.motorMaxPitch = jbeamData.motorMaxPitch or 1
    M.motorInverterEffortVolumeCoef = jbeamData.motorInverterEffortVolumeCoef or 0
    
    -- Node IDs
    if jbeamData.inverterSoundNode_nodes and type(jbeamData.inverterSoundNode_nodes) == "table" then
        M.inverterSoundNodeMG1 = jbeamData.inverterSoundNode_nodes[1]
        M.inverterSoundNodeMG2 = jbeamData.inverterSoundNode_nodes[1]
    end
    
    if jbeamData.inverterSoundNodeMG1_nodes and type(jbeamData.inverterSoundNodeMG1_nodes) == "table" then
        M.inverterSoundNodeMG1 = jbeamData.inverterSoundNodeMG1_nodes[1]
    end
    
    if jbeamData.inverterSoundNodeMG2_nodes and type(jbeamData.inverterSoundNodeMG2_nodes) == "table" then
        M.inverterSoundNodeMG2 = jbeamData.inverterSoundNodeMG2_nodes[1]
    end
    
    if jbeamData.motorSoundNode_nodes and type(jbeamData.motorSoundNode_nodes) == "table" then
        M.motorSoundNode = jbeamData.motorSoundNode_nodes[1]
    end
    
    M.startupSampleName = jbeamData.startupSampleName or "file:>vehicles>hybrid_common>sounds>hybrid_startup.ogg"
    M.startupSoundBaseVolume = jbeamData.startupSoundBaseVolume or 0.25
    M.shutdownSampleName = jbeamData.shutdownSampleName or "file:>vehicles>hybrid_common>sounds>hybrid_shutdown.ogg"
    M.shutdownSoundBaseVolume = jbeamData.shutdownSoundBaseVolume or 0.25
    
    -- Charge power curves
    M.maxChargePowerVsSpeed = simpleCurve.createSimpleCurve(jbeamData.maxChargePowerVsSpeed, "speed", "maxChargePower", {
        {"speed", "maxChargePower"},
        {15, 0},
        {45, 15},
        {60, 30},
    })
    M.maxChargePowerCoefThrottle = simpleCurve.createSimpleCurve(jbeamData.maxChargePowerCoefVsThrottle, "throttle", "coef", {
        {"throttle", "coef"},
        {85, 1},
        {100, 0},
    })
    
    -- Throttle vs RPM curve
    M.engineRpmVsTorque = simpleCurve.createSimpleCurve(jbeamData.engineRpmVsTorque, "torque", "rpm", {
        {"torque", "rpm"},
        {60, 750},
        {187, 3950},
    })
    
    -- Priority power vs speed curve
    M.maxEvOnlyPriorityPowerSpeed = simpleCurve.createSimpleCurve(jbeamData.maxEvOnlyPriorityPowerSpeed, "speed", "power", {
        {"speed", "power"},
        {0, 30},
        {25, 15},
    })
    M.maxEngineAssistPriorityPowerSpeed = simpleCurve.createSimpleCurve(jbeamData.maxEngineAssistPriorityPowerSpeed, "speed", "power", {
        {"speed", "power"},
        {0, 15},
        {25, 10},
    })
    
    -- Priority power vs battery curve
    M.maxPriorityPowerBatteryLevel = simpleCurve.createSimpleCurve(jbeamData.maxPriorityPowerBatteryLevel, "batteryLevel", "power", {
        {"batteryLevel", "power"},
        {45, 0},
        {55, 50},
    })
    
    -- PIDs
    M.driveThrottlePid = pid.newPid(
        jbeamData.throttlePidPFactor or 0.4,
        jbeamData.throttlePidIFactor or 0.85,
        jbeamData.throttlePidDFactor or 0.002
    ):setScale(1 / (jbeamData.throttlePidScale or 1000)):clampOutput(0, 1):setDynamicILock(0.6, false):setName("throttle")--:enableDebug()
    M.chargeThrottlePid = pid.newPid(
        jbeamData.chargeThrottlePidPFactor or 5.5, 
        jbeamData.chargeThrottlePidIFactor or 8.0
    ):setScale(1 / 1000):clampOutput(0, 1)

    -- Lots of temporal smoothers!
    M.throttleSmoother = newTemporalSmoothingNonLinear(
        jbeamData.throttleSmoothingIn or 1.5,
        jbeamData.throttleSmoothingOut or 1.5)
    M.throttleLimiterSmoother = newTemporalSmoothingNonLinear(jbeamData.throttleLimitSmoothing or 0.95)
    M.engineStarterSmoother = newTemporalSmoothingNonLinear(
        jbeamData.engineStarterSmoothingIn or 30,
        jbeamData.engineStarterSmoothingOut or 30)
    M.chargePowerSmoother = newTemporalSmoothingNonLinear(jbeamData.chargePowerSmoothing or 4)
    M.mg1MaxPowerSmoother = newTemporalSmoothingNonLinear(jbeamData.motorPowerSmoothing or 8.5)
    M.mg2MaxPowerSmoother = newTemporalSmoothingNonLinear(jbeamData.motorPowerSmoothing or 8.5)
        
    -- Sound smoothers
    M.mg1PitchSmoother = newTemporalSmoothingNonLinear(6)
    M.mg2PitchSmoother = newTemporalSmoothingNonLinear(6)
    
    -- Latches
    M.latchDriveModeZero = latch.newTimedLatch(jbeamData.modeZeroLatchTime or 0.25)
    M.latchDriveModeOne  = latch.newTimedLatch(jbeamData.modeOneLatchTime or 0.25, true)
    M.latchBatteryChargeUp = latch.newThresholdLatch(M.batteryChargeLowRatio, M.batteryChargeHighRatio)

    -- Pre-fill debug table with empty strings to set up proper order
    debugutil.clearValues()
    debugutil.putValue("Power", "")
    debugutil.putValue("State", "")
    debugutil.putValue("eCVT Drive Mode", "")
    debugutil.putValue("Engine Run Time", "")
    debugutil.putValue("Engine demanded on", "")
    debugutil.putValue("Engine on reason", "")
    debugutil.putValue("Engine triggered on by", "")
    debugutil.putValue("Battery power (kW)", "")
    debugutil.putValue("Generator power (kW)", "")
    debugutil.putValue("Battery SOC %", "")
    debugutil.putValue("Battery SOC % (usable)", "")
    debugutil.putValue("Input Throttle", "")
    debugutil.putValue("Throttle % EV thresh.", "")
    debugutil.putValue("Throttle Aggr.", "")
    debugutil.putValue("Requested Braking Torque", "")
    debugutil.putValue("Requested Regen Torque (Out, Wheels)", "")
    debugutil.putValue("Max Regen Torque (Out, Wheels)", "")
    debugutil.putValue("Actual Regen Torque (Out, Wheels)", "")
    debugutil.putValue("Friction Braking Torque (Req, Act)", "")
    debugutil.putValue("Total Braking Torque (%)", "")
    debugutil.putValue("Regen % of brake demand", "")
    debugutil.putValue("Friction % of full brakes", "")
    debugutil.putValue("Adjusted friction torque", "")
    debugutil.putValue("Available Power (Priority/Secondary)", "")
    debugutil.putValue("Available EV Torque (Priority/Secondary)", "")
    debugutil.putValue("Requested Torque (Total, EV, ICE)", "")
    debugutil.putValue("Actual Torque (Total, EV, ICE)", "")
    debugutil.putValue("Achievable Engine Torque (Eng, S, C, R)", "")
    debugutil.putValue("Requested Engine Torque (Eng, S, C, R)", "")
    debugutil.putValue("Actual Engine Torque (Eng, S, C, R)", "")
    debugutil.putValue("ICE HP", "")
    debugutil.putValue("ICE Torque", "")
    debugutil.putValue("Combustion Torque", "")
    debugutil.putValue("Requested Charge Power", "")
    debugutil.putValue("Requested Charge Power", "")
    debugutil.putValue("Min RPM", "")
    debugutil.putValue("Max RPM", "")
    debugutil.putValue("Min Target RPM", "")
    debugutil.putValue("Max Target RPM", "")
    debugutil.putValue("Target RPM", "")
    debugutil.putValue("Actual RPM", "")
    debugutil.putValue("RPM Error", "")
    debugutil.putValue("Load Bias", "")
    debugutil.putValue("Engine Load Torque", "")
    debugutil.putValue("Engine Deload Torque", "")
    debugutil.putValue("Charge Throttle", "")
    debugutil.putValue("Desired ICE Throttle", "")
    debugutil.putValue("Actual ICE Throttle", "")
    debugutil.putValue("ICE Load", "")
    debugutil.putValue("MG1 Max Torque Carrier AV", "")
    debugutil.putValue("MG1 Max Torque RPM", "")
    debugutil.putValue("Available MG1 Power", "")
    debugutil.putValue("Available MG2 Power", "")
    debugutil.putValue("MG2 Adjustment", "")
    debugutil.putValue("Coolant Temp.", "")
end

local function reset()
    M.init(M.jbeamData)
    M.initSecondStage()
end

local function initSounds()
    soundSources.mg1Whine = obj:createSFXSource(
        M.inverterSampleNameMG1,
        "AudioDefaultLoop3D",
        "mg1Whine",
        M.inverterSoundNodeMG1 or gearbox.transmissionNodeID
    )
    
    soundSources.mg2Whine = obj:createSFXSource(
        M.inverterSampleNameMG2,
        "AudioDefaultLoop3D",
        "mg2Whine",
        M.inverterSoundNodeMG2 or gearbox.transmissionNodeID
    )

    soundSources.motorRamp = obj:createSFXSource(
        M.motorSampleName,
        "AudioDefaultLoop3D",
        "motorRamp",
        M.motorSoundNode or gearbox.transmissionNodeID
    )
end

local function overrideSetStarter(enabled)
    if enabled then
        M.powerSwitch = not M.powerSwitch
    end
end

local function overrideSetIgnition(enabled)
    if not enabled then
        M.powerSwitch = false
    end
end

local function initSecondStage()
    engine = powertrain.getDevice("mainEngine")
    gearbox = powertrain.getDevice("gearbox")
    battery = energyStorage.getStorage("tractionBattery")
    
    local hasError = false
    
    if not engine then
        debugutil.log("E", "Engine not found; cannot continue")
        hasError = true
    end
    if not gearbox then
        debugutil.log("E", "Gearbox not found; cannot continue")
        hasError = true
    end
    if not battery then
        debugutil.log("E", "Battery not found; cannot continue")
        hasError = true
    end
    if gearbox.type ~= "ecvtGearbox" then
        debugutil.log("E", "Gearbox must be of type ecvtGearbox; cannot continue")
        hasError = true
    end
    
    if hasError then
        M.updateGFX = nop
        return
    end
    
    if M.jbeamData.turboRegen and v.data[M.jbeamData.turboRegen] then
        local turboRegenFileName = M.jbeamData.turboRegenLuaFileName or "powertrain/turboRegen"
        local turboRegenModule = require(turboRegenFileName)
        
        M.turboRegen = turboRegenModule.new(v.data[M.jbeamData.turboRegen])
        
        if not M.turboRegen then
            error("Error initializing turbo pre-spooler; check logs for details")
        end
    else
        M.turboRegen = { updateGFX = nop, reset = nop, isExisting = false }
    end
    
    -- Override setStarter and setEngineIgnition
    if controller.mainController then
        controller.mainController.setStarter = overrideSetStarter
        controller.mainController.setEngineIgnition = overrideSetIgnition
    end

    M.minBatteryPowerLevel = battery.jbeamData.minWorkingSoc or M.minBatteryPowerLevel
    M.maxBatteryPowerLevel = battery.jbeamData.maxWorkingSoc or M.maxBatteryPowerLevel
    M.criticalBatteryPowerLevel = battery.jbeamData.criticalSoc or M.criticalBatteryPowerLevel

    if not M.maxFastChargePower then
        M.maxFastChargePower = battery.jbeamData.maxPowerInput
    end
    
    engineTorqueData = engine:getTorqueData()
    
    if not M.jbeamData.engineMaxRpm then
        M.engineMaxRpm = engineTorqueData.maxPowerRPM
    end
    
    if not M.engineMinIdleRpm then
        M.engineMinIdleRpm = engine.idleRPM + 200
    end

    calculateTorqueStatistics()
    printTorqueData()

    totalPossibleBrakeTorque = 0
    totalPropulsedBrakeTorque = 0
    totalUnpropulsedBrakeTorque = 0
    
    for i = 0, wheels.wheelCount - 1, 1 do
        local wd = wheels.wheels[i]
        
        totalPossibleBrakeTorque = totalPossibleBrakeTorque + wd.brakeTorque
        
        if wd.isPropulsed then
            totalPropulsedBrakeTorque = totalPropulsedBrakeTorque + wd.brakeTorque
        else
            totalUnpropulsedBrakeTorque = totalUnpropulsedBrakeTorque + wd.brakeTorque
        end
        
        if not initialWheelBrakeSplit[i] then
            initialWheelBrakeSplit[i] = {wd.brakeInputSplit, wd.brakeSplitCoef}
        end
    end

    if M.startEvMode then
        electrics.values.evMode = 1
    end
    
    if M.startPoweredOff then
        engine.outputAV = 0
    end
    
    timeSinceInit = max(0, M.initializeTime - 0.25)
    state = "init"
    nextState = nil
    
    M.chargeThrottlePid:reset()
    M.driveThrottlePid:reset()
    M.latchDriveModeZero:reset()
    M.latchDriveModeOne:reset()
end

M.updateGFX = updateGFX
M.init = init
M.initSounds = initSounds
M.initSecondStage = initSecondStage
M.reset = reset
M.getTorqueData = getTorqueData
M.printTorqueData = printTorqueData

return M
